@@ -74,10 +74,18 @@
 
             labels <- labels[!is.na(labels) | tagged]
             if (length(labels) > 0) {
-                names(labels) <- cleanup(names(labels))
-                result[["labels"]] <- labels
+                # names(labels) <- cleanup(names(labels))
+                result[["labels"]] <- setNames(labels, cleanup(names(labels)))
             }
         }
+        else if (is.factor(x)) {
+            xlevels <- levels(x)
+            # labels <- seq(length(xlevels))
+            # names(labels) <- xlevels
+            # result[["labels"]] <- labels
+            result[["labels"]] <- setNames(seq(length(xlevels)), xlevels)
+            x <- as.numeric(x)
+        }
         
         na_values <- attr(x, "na_values", exact = TRUE)
         if (is.null(na_values)) {
@@ -98,12 +106,24 @@
         }
         
         result$na_range <- attr(x, "na_range", exact = TRUE)
-        result$type <- DDIwR::checkType(
+        result$type <- checkType(
             x,
             labels,
-            admisc::possibleNumeric(x)
+            na_values
         )
 
+        format.spss <- attr(x, "format.spss", exact = TRUE)
+        if (is.null(format.spss)) {
+            format.spss <- getFormat(x, type = "SPSS")
+        }
+
+        format.stata <- attr(x, "format.stata", exact = TRUE)
+        if (is.null(format.stata)) {
+            format.stata <- getFormat(x, type = "Stata")
+        }
+
+        result[["varFormat"]] <- c(format.spss, format.stata)
+        
         return(result)
     })
 
