@@ -24,7 +24,7 @@
 # SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 `getMetadata` <- 
-function(x, save = FALSE, OS = "Windows", ...) {
+function(x, save = FALSE, declared = TRUE, OS = "Windows", encoding = "UTF-8", ...) {
     
     # TODO: detect DDI version or ask the version through a dedicated argument
     # http://www.ddialliance.org/Specification/DDI-Codebook/2.5/XMLSchema/field_level_documentation.html
@@ -38,14 +38,8 @@ function(x, save = FALSE, OS = "Windows", ...) {
     ) {
         user_na <- dots$user_na
     }
-
-    encoding <- "latin1"
-    if (
-        is.element("encoding", names(dots)) && is.atomic(dots$encoding) && 
-        length(dots$encoding) == 1 && is.character(dots$encoding)
-    ) {
-        encoding <- dots$encoding
-    }
+    
+    embed <- isTRUE(dots$embed)
     
     if (is.data.frame(x)) {
         error <- TRUE
@@ -68,24 +62,20 @@ function(x, save = FALSE, OS = "Windows", ...) {
         else {
             codeBook <- list()
             codeBook$dataDscr <- collectMetadata(x)
-            
+            if (embed) {
+                codeBook$fileDscr <- list(
+                    datafile = x
+                )
+            }
             return(codeBook)
         }
     }
     
     enter <- getEnter(OS)
     
-    fromsetupfile <- FALSE
-    if (is.element("fromsetupfile", names(dots))) {
-        fromsetupfile <- dots$fromsetupfile
-    }
+    fromsetupfile <- isTRUE(dots$fromsetupfile)
     
-    embed <- FALSE
-    if (is.element("embed", names(dots))) {
-        embed <- dots$embed
-    }
-    
-    tp <- DDIwR::treatPath(x, type = "*")
+    tp <- treatPath(x, type = "*")
     
     singlefile <- length(tp$files) == 1
     notes <- NA
@@ -170,7 +160,9 @@ function(x, save = FALSE, OS = "Windows", ...) {
                 values <- cleanup(xml2::xml_text(xml2::xml_find_all(vars[i], xpath)))
 
                 xpath <- sprintf("%svarFormat", dns)
-                type <- xml2::xml_attr(xml2::xml_find_first(vars[i], xpath), "type")
+                vformat <- xml2::xml_find_first(vars[i], xpath)
+                type <- xml2::xml_attr(vformat, "type")
+                varFormat <- xml2::xml_text(vformat)
                 
                 if (length(values) > 0) {
                     
@@ -216,20 +208,7 @@ function(x, save = FALSE, OS = "Windows", ...) {
                     codeBook$dataDscr[[i]]$na_range <- na_range
                 }
 
-                if (!is.na(measurement)) {
-                    if (is.element(measurement, c("nominal", "ordinal"))) {
-                        codeBook$dataDscr[[i]]$type <- "cat"
-                    }
-                    else if (is.element(measurement, c("interval", "ratio"))) {
-                        codeBook$dataDscr[[i]]$type <- "num"
-                    }
-                    else if (!is.na(type)) {
-                        codeBook$dataDscr[[i]]$type <- type
-                    }
-
-                    codeBook$dataDscr[[i]]$measurement <- measurement
-                }
-                else {
+                if (is.na(measurement)) {
                     if (!is.na(type)) {
                         codeBook$dataDscr[[i]]$type <- "num" # default
 
@@ -243,6 +222,23 @@ function(x, save = FALSE, OS = "Windows", ...) {
                         }
                     }
                 }
+                else {
+                    if (is.element(measurement, c("nominal", "ordinal"))) {
+                        codeBook$dataDscr[[i]]$type <- "cat"
+                    }
+                    else if (is.element(measurement, c("interval", "ratio"))) {
+                        codeBook$dataDscr[[i]]$type <- "num"
+                    }
+                    else if (!is.na(type)) {
+                        codeBook$dataDscr[[i]]$type <- type
+                    }
+
+                    codeBook$dataDscr[[i]]$measurement <- measurement
+                }
+
+                if (!is.na(vformat)) {
+                    codeBook$dataDscr[[i]]$varFormat <- varFormat
+                }
 
                 if (identical(type, "character")) {
                     xpath <- sprintf("%stxt", dns)
@@ -305,7 +301,13 @@ function(x, save = FALSE, OS = "Windows", ...) {
                 indent <- dots$indent
             }
             
-            writeRlist(codeBook$dataDscr, OS = OS, indent = indent, dirpath = tp$completePath, filename = tp$filenames[ff])
+            writeRlist(
+                codeBook$dataDscr,
+                OS = OS,
+                indent = indent,
+                dirpath = tp$completePath,
+                filename = tp$filenames[ff]
+            )
             
         }
     }
@@ -315,11 +317,26 @@ function(x, save = FALSE, OS = "Windows", ...) {
     if (singlefile) {
         if (!is.na(notes)) {
             if (grepl("# start data #", notes)) {
-                spss <- ifelse(is.element("spss", names(dots)), dots$spss, TRUE)
+                # this can only be possible from an XML, DDI Codebook
+                # therefore the varFormat should always be of an SPSS type
                 notes <- unlist(strsplit(notes, split = "\\n"))
-                data <- notes[seq(which(grepl("# start data #", notes)) + 1, which(grepl("# end data #", notes)) - 1)]
+                data <- notes[
+                    seq(
+                        which(grepl("# start data #", notes)) + 1,
+                        which(grepl("# end data #", notes)) - 1
+                    )
+                ]
+
                 data <- read.csv(text = paste(data, collapse = "\n"), as.is = TRUE)
-                data <- make_labelled(data, codeBook$dataDscr, spss = spss)
+                # return(list(data, codeBook$dataDscr, declared = declared, spss = spss))
+
+                # make_labelled is always and only about SPSS type of variables
+                data <- make_labelled(
+                    data,
+                    codeBook$dataDscr,
+                    declared = declared
+                )
+
                 embed <- TRUE
             }
         }
