@@ -23,10 +23,8 @@
 # (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 # SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
-`make_labelled` <- function(x, dataDscr, ...) {
+`make_labelled` <- function(x, dataDscr, declared = TRUE) {
 
-    pN <- sapply(x, admisc::possibleNumeric)
-    
     for (i in names(x)) {
         #------------------------------------------------------------------
         # attrx$label, if not existing, takes from attrx$labels
@@ -38,23 +36,67 @@
         na_values <- dataDscr[[i]][["na_values"]]
         na_range <- dataDscr[[i]][["na_range"]]
 
-        v <- unname(unlist(unclass(x[, i])))
-        if (pN[i]) {
+        v <- x[[i]]
+        attributes(v) <- NULL
+
+        pN <- TRUE
+        allnav <- all(is.na(v))
+        nullabels <- is.null(labels)
+        if (!(allnav & nullabels)) {
+            pN <- admisc::possibleNumeric(c(v, unname(labels)))
+        }
+
+        if (pN) {
             v <- admisc::asNumeric(v)
         }
+        else {
+            v <- as.character(v)
+            na_range <- NULL
+        }
+
+        if (!nullabels) {
+            nms <- names(labels)
+            if (pN) {
+                labels <- setNames(admisc::asNumeric(labels), nms)
+            }
+            else {
+                labels <- setNames(as.character(labels), nms)
+            }
+            # names(labels) <- nms
+        }
+
+        if (!is.null(na_values)) {
+            if (admisc::possibleNumeric(na_values) & pN) {
+                na_values <- admisc::asNumeric(na_values)
+            }
+            else {
+                na_values <- as.character(na_values)
+            }
+        }
         
-        x[[i]] <- declared::declared(v, labels, na_values, na_range, label)
+        if (declared) {
+            x[[i]] <- declared::declared(v, labels, na_values, na_range, label)
+        }
+        else {
+            x[[i]] <- haven::labelled_spss(v, labels, na_values, na_range, label)
+        }
+
+        # this is always about format.spss since both "declared" and "labelled_spss"
+        # are not using Stata type extended missing values, and by consequence
+        # not using the Stata format type
+        attr(x[[i]], "format.spss") <- dataDscr[[i]][["varFormat"]][1]
 
     }
 
-    other.args <- list(...)
-    if (is.element("spss", names(other.args))) {
-        if (other.args$spss) {
-            x[] <- lapply(x, function(x) {
-                attr(x, "format.spss") <- getFormat(x)
-                return(x)
-            })
+    x[] <- lapply(x, function(x) {
+        if (is.null(attr(x, "format.spss"))) {
+            attr(x, "format.spss") <- getFormat(x, type = "SPSS")
         }
+        return(x)
+    })
+
+    if (!declared) {
+        class(x) <- c("tbl_df", "tbl", "data.frame")
     }
 
     return(x)
