@@ -24,23 +24,16 @@
 # SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 `recodeValues` <- function(
-    dataset, to = c("SPSS", "Stata"), dictionary = NULL, chartonum = TRUE, ...
+    dataset, to = c("SPSS", "Stata"), dictionary = NULL, ...
 ) {
 
     to <- toupper(match.arg(to))
     too_many <- FALSE
-    to_declared <- TRUE
-
+    
     dots <- list(...)
 
-    error_null <- TRUE
-    if (is.element("error_null", names(dots))) {
-        error_null <- dots$error_null
-    }
-
-    if (is.element("to_declared", names(dots))) {
-        to_declared <- dots$to_declared
-    }
+    error_null <- ifelse(isFALSE(dots$error_null), FALSE, TRUE)
+    to_declared <- ifelse(isFALSE(dots$to_declared), FALSE, TRUE)
 
     if (is.data.frame(dataset)) {
         error <- TRUE
@@ -78,18 +71,24 @@
     dataDscr <- collectMetadata(dataset, error_null = error_null)
 
     spss <- unlist(lapply(dataset, function(x) {
-        !is.null(attr(x, "labels")) &&
+        # it makes sense to check for character variables, since
+        # neither Stata nor SAS do not accept missing values for chars
+        # technically, a char var with missing value would be "valid" in SPSS
+        # but it doesn't matter if recoding to Stata, it's like it would not exist
+        !is.character(x) &&
+        !is.null(attr(x, "labels", exact = TRUE)) &&
         (
             inherits(x, "haven_labelled_spss") || inherits(x, "declared")
         )
     }))
 
     if ((sum(spss) == 0 & to == "STATA") | (sum(spss) == ncol(dataset) & to == "SPSS")) {
+        if (isTRUE(dots$return_dictionary)) {
+            return(NULL)
+        }
         return(dataset)
     }
 
-    # build a dictionary based on existing metadata
-
     allMissing <- list()
     
     for (variable in names(dataset)) {
@@ -97,30 +96,8 @@
         attrx <- attributes(x)
         attributes(x) <- NULL
 
-        if (is.element("declared", attrx$class) & to == "STATA") {
-            # transform back to a haven_labelled variable
-            na_index <- attrx$na_index
-            
-            if (!is.null(na_index)) {
-                nms <- names(na_index)
-                if (admisc::possibleNumeric(nms) || all(is.na(nms))) {
-                    nms <- admisc::asNumeric(nms)
-                    if (admisc::wholeNumeric(nms)) {
-                        nms <- as.integer(nms)
-                    }
-                }
-                
-                x[na_index] <- nms
-            }
-            
-            attrx$class <- c("haven_labelled_spss", "haven_labelled", "vctrs_vctr", class(x))
-
-            dataset[[variable]] <- x
-            attrx$na_index <- NULL
-            attributes(dataset[[variable]]) <- attrx
-        }
-
         metadata <- dataDscr[[variable]]
+        labels <- metadata[["labels"]]
         missing <- NULL
 
         if (is.element("na_values", names(metadata))) {
@@ -130,8 +107,21 @@
         if (is.element("na_range", names(metadata))) {
             na_range <- metadata$na_range
             misvals <- x[x >= na_range[1] & x <= na_range[2]]
-            missing <- sort(unique(c(missing, misvals[!is.na(misvals)])))
+            missing <- c(missing, misvals[!is.na(misvals)], na_range)
+
+            if (!is.null(labels)) {
+                if (admisc::possibleNumeric(labels)) {
+                    labels <- admisc::asNumeric(labels)
+                    missing <- c(
+                        missing,
+                        labels[labels >= na_range[1] & labels <= na_range[2]]
+                    )
+                }
+            }
+
+            missing <- sort(unique(missing))
         }
+
         if (
             is.element("labels", names(metadata)) &&
             any(is.element(missing, metadata[["labels"]]))
@@ -140,12 +130,12 @@
             names(missing)[wel] <- names(metadata[["labels"]])[
                 match(missing[wel], metadata[["labels"]])
             ]
+            missing <- missing[wel]
         }
 
         allMissing[[variable]] <- missing
 
     }
-# return(allMissing)
 
     missingSPSS <- missingStata <- NULL
 
@@ -157,13 +147,14 @@
         umis <- unique(data.frame(labels = names(umis), codes = umis))
         
         if (nrow(umis) == 0) {
-            admisc::stopError("There is no information about categories, labels and missing values")
+            # There is no information about missing values
+            return(dataset)
         }
 
         umis <- umis[order(umis$labels, umis$codes), ]
         umis$rec <- umis$codes
         ulabels <- unique(umis[, 1])
-
+        
         for (i in seq(length(ulabels))) {
             if (ulabels[i] != "") {
                 uel <- is.element(umis$labels, ulabels[i])
@@ -178,9 +169,9 @@
             wneg <- which(neg)[order(umis$rec[neg], decreasing = TRUE)]
         }
         umis <- umis[c(wneg, which(!neg)), ]
-# print(umis)
-        missingSPSS <- umis$codes
-        names(missingSPSS) <- umis$rec
+        # print(umis)
+        missingSPSS <- setNames(umis$codes, umis$rec)
+        # names(missingSPSS) <- umis$rec
         
         # missingSPSS <- sort(unique(unname(unlist(allMissing[spss]))))
     }
@@ -191,19 +182,30 @@
 
     if (to == "SPSS") {
         torecode <- missingStata
-        temp <- undeclare(dataset)
-        all_neg <- unique(temp[temp < 0])
-        attributes(all_neg) <- NULL
+        # return(dataset)
+        temp <- unlist(lapply(undeclare(dataset), function(x) {
+            attributes(x) <- NULL
+            if (is.character(x)) return(NA)
+            return(x[x < 0])
+        }))
+        # temp <- unlist(lapply(dataset, function(x) {
+        #     attributes(x) <- NULL
+        #     if (is.character(x)) return(NA)
+        #     return(x[x < 0])
+        # }))
+        attributes(temp) <- NULL
+        all_neg <- unique(temp)
         
-        if (length(all_neg) > 0) {
+        if (!identical(all_neg, NA) && length(all_neg) > 0) {
             all_neg <- all_neg[!is.na(all_neg)]
         }
+
         if (!is.null(torecode)) {
             toreplace <- 90 + seq(length(missingStata) + length(all_neg) + length(missingSPSS))
             toreplace <- -1 * setdiff(toreplace, missingSPSS)
             toreplace <- setdiff(toreplace, all_neg)
-            torecode <- toreplace[seq(length(missingStata))]
-            names(torecode) <- missingStata
+            torecode <- setNames(toreplace[seq(length(missingStata))], missingStata)
+            # names(torecode) <- missingStata
         }
     }
     else if (to == "STATA") {
@@ -238,6 +240,35 @@
 
     if (is.null(dictionary)) {
         dictionary <- torecode
+        nms <- names(dictionary)
+        if (any(duplicated(dictionary))) {
+            for (i in which(duplicated(dictionary))) {
+                first <- which(dictionary == dictionary[i])[1]
+                nms[nms == nms[i]] <- nms[first]
+            }
+            names(dictionary) <- nms
+            dfd <- data.frame(nms = nms, vals = dictionary)
+            dfd <- unique(dfd)
+            dfd <- dfd[order(dfd$vals, dfd$nms), ]
+            dictionary <- setNames(dfd$vals, dfd$nms)
+            # names(dictionary) <- dfd$nms
+        }
+
+        unms <- unique(nms)
+        if (all(is.element(unms, letters))) {
+            for (i in seq(length(unms))) {
+                nms[nms == unms[i]] <- letters[i]
+            }
+        }
+        else {
+            for (i in seq(length(unms))) {
+                nms[nms == unms[i]] <- -90 - i
+            }
+        }
+        
+        if (isTRUE(dots$return_dictionary)) {
+            return(dictionary)
+        }
     }
     else if (!is.null(torecode)) {
         if (to == "SPSS") {
@@ -262,181 +293,151 @@
     }
 
     for (i in seq(ncol(dataset))) {
-        x <- unclass(dataset[[i]])
+        # print(i)
+        x <- unclass(declared::undeclare(dataset[[i]]))
         metadata <- dataDscr[[i]]
         labels <- metadata[["labels"]]
         na_values_i <- metadata[["na_values"]]
         na_range_i <- metadata[["na_range"]]
 
-        if (to == "SPSS") {
-            if (!spss[i] && !is.null(dictionary)) {
+        if (!is.null(na_values_i) | !is.null(na_range_i)) {
+            if (to == "SPSS") {
+                if (!spss[i] && !is.null(dictionary)) {
 
-                na_values_i <- na_values[is.element(nms, metadata[["na_values"]])]
-                nms_i <- nms[is.element(nms, metadata[["na_values"]])]
+                    na_values_i <- na_values[is.element(nms, metadata[["na_values"]])]
+                    nms_i <- nms[is.element(nms, metadata[["na_values"]])]
 
-                # if (i == 10) print(na_values_i)
-                
-                if (length(na_values_i) > 0) {
-                    for (d in seq(length(na_values_i))) {
-                        if (nchar(nms[d]) == 1) {
-                            x[haven::is_tagged_na(x, nms_i[d])] <- na_values_i[d]
-                            labels[haven::is_tagged_na(labels, nms_i[d])] <- na_values_i[d]
+                    # if (i == 10) print(na_values_i)
+                    if (length(na_values_i) > 0) {
+                        for (d in seq(length(na_values_i))) {
+                            if (nchar(nms[d]) == 1) {
+                                x[haven::is_tagged_na(x, nms_i[d])] <- na_values_i[d]
+                                labels[haven::is_tagged_na(labels, nms_i[d])] <- na_values_i[d]
+                            }
                         }
                     }
                 }
-            }
 
-            if (!admisc::possibleNumeric(labels)) {
-                x <- as.character(x)
-                if (length(na_values_i) > 0) {
-                    na_values_i <- as.character(na_values_i)
-                }
-            }
-
-            # cat(
-            #     paste(
-            #         i,
-            #         "--",
-            #         paste(
-            #             na_values,
-            #             collapse = ","
-            #         ),
-            #         "--",
-            #         paste(
-            #             na_values_i,
-            #             collapse = ","
-            #         ),
-            #         "--",
-            #         paste(
-            #             metadata[["na_values"]],
-            #             collapse = ","
-            #         ),
-            #         "\n"
-            #     )
-            # )
-
-            callist <- list(
-                x = x,
-                labels = labels,
-                label = metadata[["label"]]
-            )
-
-            if (length(na_values_i) > 0) {
-                if (is.character(na_values_i) && length(na_values_i) > 3) {
-                    na_values_i <- na_values_i[1:3]
-                }
-
-                if (length(na_values_i) > 3) {
-                    callist$na_range <- sort(range(na_values_i))
-                }
-                else {
-                    callist$na_values <- na_values_i
-                }
-            }
-
-            # print(callist)
-            # cat("====================\n")
-            if (to_declared) {
-                dataset[[i]] <- do.call(declared::declared, callist)
-            }
-            else {
-                dataset[[i]] <- do.call(haven::labelled_spss, callist)
-            }
-            
-        }
-        else if (to == "STATA") {
-            
-            if (spss[i] & !is.null(dictionary)) {
-                attributes(x) <- NULL
-
-                ux <- unique(c(x[!is.na(x)], labels))
-                pnux <- admisc::possibleNumeric(ux, each = TRUE)
-                nums <- c()
-                if (any(pnux)) {
-                    nums <- as.numeric(ux[pnux])
-                }
-
-                # transform character values and labels to numeric
-                # Stata does not allow character values
-                if (any(!pnux) & chartonum & !is.null(labels)) {
-                    nms_l <- names(labels)
-
-                    cux <- ux[!pnux]
-                    lux <- length(cux)
-                    n <- l <- 1
-                    
-                    while (l <= lux) {
-                        if (is.element(n, nums)) {
-                            n <- n + 1
-                        }
-                        else {
-                            x[x == cux[l]] <- n
-                            labels[labels == cux[l]] <- n
-                            n <- n + 1
-                            l <- l + 1
-                        }
+                if (!is.null(labels) && !admisc::possibleNumeric(labels)) {
+                    x <- as.character(x)
+                    if (length(na_values_i) > 0) {
+                        na_values_i <- as.character(na_values_i)
                     }
-
-                    labels <- as.numeric(labels)
-                    names(labels) <- nms_l
-                }
-                
-                selection <- rep(TRUE, length(dictionary))
-
-                if (!is.null(na_values_i)) {
-                    selection <- is.element(dictionary, na_values_i)
-                }
-                else if (!is.null(na_range_i)) {
-                    selection <- as.numeric(dictionary) >= na_range_i[1] & as.numeric(dictionary) <= na_range_i[2]
                 }
 
-                dic_i <- dictionary[selection]
-                
-                if (any(duplicated(dic_i))) {
-                    # multiple missing labels with the same code
-                    selabels <- umis$labels[selection]
+                # cat(
+                #     paste(
+                #         i,
+                #         "--",
+                #         paste(
+                #             na_values,
+                #             collapse = ","
+                #         ),
+                #         "--",
+                #         paste(
+                #             na_values_i,
+                #             collapse = ","
+                #         ),
+                #         "--",
+                #         paste(
+                #             metadata[["na_values"]],
+                #             collapse = ","
+                #         ),
+                #         "\n"
+                #     )
+                # )
 
-                    if (!is.null(na_values[i])) {
-                        na_labels <- names(labels)[
-                            is.element(
-                                labels,
-                                if (is.numeric(labels)) as.numeric(na_values_i) else na_values_i
-                            )
-                        ]
-                    }
-                    else if (!is.null(na_range_i)) {
-                        na_labels <- names(labels)[
-                            as.numeric(labels) >= na_range_i[1] & as.numeric(labels) <= na_range_i[2]
-                        ]
-                    }
-                    
-                    dic_i <- dic_i[is.element(selabels, na_labels)]
-                }
-                
-                nms_i <- names(dic_i)
-                if (admisc::possibleNumeric(dic_i)) {
-                    dic_i <- admisc::asNumeric(dic_i)
-                }
-
-                x <- as.numeric(x)
-                for (d in seq(length(dic_i))) {
-                    x[is.element(x, dic_i[d])] <- haven::tagged_na(nms_i[d])
-                    labels[
-                        is.element(labels, dic_i[d])
-                    ] <- haven::tagged_na(nms_i[d])
-                }
-
-                dataset[, i] <- haven::labelled(
-                    x,
+                callist <- list(
+                    x = x,
                     labels = labels,
                     label = metadata[["label"]]
                 )
+
+                if (length(na_values_i) > 0) {
+                    if (is.character(na_values_i) && length(na_values_i) > 3) {
+                        na_values_i <- na_values_i[1:3]
+                    }
+
+                    if (length(na_values_i) > 3) {
+                        callist$na_range <- sort(range(na_values_i))
+                    }
+                    else {
+                        callist$na_values <- na_values_i
+                    }
+                }
+
+                if (to_declared) {
+                    dataset[[i]] <- do.call(declared::declared, callist)
+                }
+                else {
+                    dataset[[i]] <- do.call(haven::labelled_spss, callist)
+                }
+                
             }
-            else {
-                if (inherits(dataset[[i]], "declared") & !to_declared) {
-                    dataset[[i]] <- x
+            else if (to == "STATA") {
+                
+                if (spss[i] & !is.null(dictionary)) {
+                    attributes(x) <- NULL
+                    
+                    selection <- rep(TRUE, length(dictionary))
+
+                    if (!is.null(na_values_i)) {
+                        selection <- is.element(dictionary, na_values_i)
+                    }
+                    else if (!is.null(na_range_i)) {
+                        selection <- as.numeric(dictionary) >= na_range_i[1] & as.numeric(dictionary) <= na_range_i[2]
+                    }
+
+                    dic_i <- dictionary[selection]
+                    
+                    if (any(duplicated(dic_i))) {
+                        # multiple missing labels with the same code
+                        selabels <- umis$labels[selection]
+
+                        if (!is.null(na_values[i])) {
+                            na_labels <- names(labels)[
+                                is.element(
+                                    labels,
+                                    if (is.numeric(labels)) as.numeric(na_values_i) else na_values_i
+                                )
+                            ]
+                        }
+                        else if (!is.null(na_range_i)) {
+                            num_labels <- as.numeric(labels)
+                            na_labels <- names(labels)[
+                                num_labels >= na_range_i[1] & num_labels <= na_range_i[2]
+                            ]
+                        }
+                        
+                        dic_i <- dic_i[is.element(selabels, na_labels)]
+                    }
+                    
+                    nms_i <- names(dic_i)
+                    if (admisc::possibleNumeric(dic_i)) {
+                        dic_i <- admisc::asNumeric(dic_i)
+                    }
+
+                    x <- admisc::asNumeric(x)
+
+                    for (d in seq(length(dic_i))) {
+                        x[is.element(x, dic_i[d])] <- haven::tagged_na(nms_i[d])
+                        labels[
+                            is.element(labels, dic_i[d])
+                        ] <- haven::tagged_na(nms_i[d])
+                    }
+                    
+                    dataset[, i] <- haven::labelled(
+                        x,
+                        labels = labels,
+                        label = metadata[["label"]]
+                    )
+                }
+                else {
+                    if (inherits(dataset[[i]], "declared") & !to_declared) {
+                        dataset[[i]] <- x
+                    }
+                    # otherwise leave it intact
                 }
-                # otherwise leave it intact
             }
         }
     }
